declare('DroidButtons', {version = '2.0.0-a4-tppfix'})

if (not gkinterface.IsTouchModeEnabled()) then
	console_print("DroidButtons not loaded. You are not playing on a device with a touch screen.")
	return
else
	console_print('DroidButtons loaded (v'..DroidButtons.version..'). HUD size is '..HUDSize(1,1))
end

dofile('ui.lua')

-- localized functions
local tinsert = table.insert
local rcreate = gkinterface.CreateTouchRegion
local rdestroy = gkinterface.DestroyTouchRegion
local rInt = gkini.ReadInt
local rStr = gkini.ReadString

local function ispluscommand(str) 
	if (type(str) == 'string') then
		if (str:match('^+')) then return true end
	end
	return false
end

-- local variables
local hud_type -- this will be set by the HUD_SHOW event and cleared by the HUD_HIDE event

-- initialize environment

local db = DroidButtons
db.notesid = 37643
db.data_version = 1
db.major = 'DroidButtons'

db.IMAGE_DIR = 'plugins/droidbuttons/images/'

db.place_button_mode = false
db.snap_to_grid = rInt(db.major, 'SnapToGrid', 1) == 1
db.grid_size = rInt(db.major, 'GridSize', 16)
db.long_press_delay = rInt(db.major, 'LongPressDelay', 1000)
db.disable_default_ui = rInt(db.major, 'DisableDefaultUI', 0) == 1
db.use_custom_accel = rInt(db.major, 'CustomAccel', 0) == 1
db.accelx = rStr(db.major, 'AccelX', 'NONE')
db.accely = rStr(db.major, 'AccelY', 'NONE')
if (db.disable_default_ui and Subplatform == "XperiaPlay") then db.disable_default_ui = false end

db.defaults = {
	button_width = rInt(db.major, 'DefaultWidth', 64),
	button_height = rInt(db.major, 'DefaultHeight', 64),
	field_width = rInt(db.major, 'DefaultFieldWidth', 200),
	field_height = rInt(db.major, 'DefaultFieldHeight', 200),
	font = rInt(db.major, 'DefaultFont', 18),
	fgcolor = '',
	bgcolor_button = '255 255 255 100 *',
	bgcolor_field = '255 255 255 255 *',
}

db.xres = gkinterface.GetXResolution()
db.yres = gkinterface.GetYResolution()
db.hudsize = HUDSize(1,1)

db.buttondefs = {} --{title=<string>, type=('button'|'field'), command1=<string>, command2=<string> [, command2cancel=<string> ], /
                   -- font=<number>, x=<number>, y=<number>, w=<number>, h=<number>[, invertX=<boolean>][, invertY=<boolean>] /
                   -- [, fgcolor=<string>][, bgcolor=<string>]}

db.buttons = {size = db.hudsize, active = "NO"} -- will hold button userdata and other cbox parameters
db.regions = {} -- will hold index keys matching Touch Regions with values pointing to the button index they represent

function db.remove_default_ui()
	HUD:DestroyAimTouchRegion()
	HUD:DestroyTouchLayer()
	-- Detach it instead of visible = 'NO', since for now it appears to be made visible again when HUD.dlg is shown.
	-- Don't destroy it, since there are other pieces of code that in the default IF that may need to reference it
	iup.Detach(HUD.touchdisplaylayer)

	-- Don't worry about saving these functions. Until ray separates the two parts of CreateTouchLayer,
	-- a ReloadInterface() is going to be necessary to fully restore functionality anyway.
	HUD.CreateAimTouchRegion = function() end
	HUD.CreateTouchLayer = function(self) 
		local def = {x=0, y=0, w=760, h=124}
		self.chattouchregion = rcreate(nil, nil, "say_channel", false, false, false, false, false, false, def.x,def.y, (def.x + def.w - 1),(def.y + def.h - 1))
	end
	HUD:CreateTouchLayer()
end

function db:set_accel_commands(init)
	if (db.use_custom_accel) then
		local xcmd = self.accelx; if (xcmd == 'NONE') then xcmd = '' end
		local ycmd = self.accely; if (ycmd == 'NONE') then ycmd = '' end
		gkinterface.BindCommand(2147483649, xcmd)
		gkinterface.BindCommand(2147483650, ycmd)
	end
end


local data_changes = {} -- Index is the data version.  Value is a function that will update the data structure from the previous version
data_changes[1] = function(data)
		for _, v in ipairs(data) do
			v.type = 'button'
		end
	end

function db:load_buttondefs()
	local saved_data = unspickle(LoadSystemNotes(db.notesid))
	if (not saved_data.version) then
		saved_data = {version = 0, buttondefs = saved_data}
	end

	--TODO give the user a chance to abort the update and not load the plugin
	
	-- Update the data structure so the code doesn't have to be riddled with exceptions.
	if (saved_data.version < db.data_version) then
		-- Create a backup of the saved data, just in case.
		local backupid = (db.notesid + saved_data.version) * -1
		local backup_data
		if (saved_data.version == 0) then backup_data = spickle(saved_data.buttondefs) else backup_data = spickle(saved_data) end
		SaveSystemNotes(backup_data, backupid)
		local msg = 'The latest version of DroidButtons has modifed the structure of it\'s saved data.\n'
		msg = msg..'DroidButtons will attempt to apply the appropriate updates without modifying your existing data.\n\n'
		msg = msg..'A backup has been created at "<Vendetta Directory>/settings/'..GetPlayerName()..'/system'..backupid..'notes.txt.'
		--OpenAlarm('DroidButtons: Update Notice', msg, 'OK')

		-- apply each incremental
		for i = saved_data.version + 1, db.data_version do
			data_changes[i](saved_data.buttondefs)
		end
		SaveSystemNotes(spickle({version = db.data_version, buttondefs = saved_data.buttondefs}), db.notesid)

	elseif (saved_data.version > db.data_version) then
		local msg = 'The version of DroidButtons you are using may not be compatible with your saved button data.\n'
		msg = msg .. 'In order to avoid errors or damaging your saved button data, you should install a newer version of DroidButtons.'
		OpenAlarm('DroidButtons: Data Version Mismatch', msg, 'OK')
	end
	db.buttondefs = saved_data.buttondefs
end

function db:save_buttondefs()
	SaveSystemNotes(spickle({version = self.data_version, buttondefs = self.buttondefs}), self.notesid)
end

function db:add_button(buttondef)
	local default = self.defaults
	-- create the button definition
	local button = {
		title = buttondef.title or '', 
		size = (buttondef.w or default.width)..'x'..(buttondef.h or default.height),
		cx = buttondef.x or 0,
		cy = buttondef.y or 0,
		active = 'YES',
		fgcolor = buttondef.fgcolor,
		bgcolor = buttondef.bgcolor or default['bgcolor_'..buttondef.type],
		centeruv = ".5 .5 .5 .5",
		font = buttondef.font or default.font,
	}

	-- create the button
	local b
	if ((buttondef.type or 'button') == 'button') then
		b = iup.stationbutton(button)
		if ((buttondef.command1 and not ispluscommand(buttondef.command1)) or buttondef.command2) then b.timer = Timer() end
	elseif (buttondef.type == 'field') then
		button.image = IMAGE_DIR..'hud_radar.png'
		b = iup.label(button)
	end
	if (b) then tinsert(self.buttons, b) end
end

function db:update_button(index, buttondef)
	--if buttondef is present, use that.  otherwise reference buttondef by index.
	buttondef = buttondef or self.buttondefs[index]
	local default = self.defaults
	local button = self.buttons[index]
	
	if (buttondef and button) then
		button.title = buttondef.title or ''
		button.cx = buttondef.x or 0
		button.cy = buttondef.y or 0
		button.size = (buttondef.w or default.width)..'x'..(buttondef.h or default.height)
		button.font = buttondef.font or default.font

		if ((buttondef.type or 'button') == 'button') then
			if ((buttondef.command1 and not ispluscommand(buttondef.command1)) or buttondef.command2) then 
				button.timer = Timer() 
			else
				button.timer = nil
			end
		end
		if (db.cbox) then iup.Refresh(db.cbox) end
	end
end

function db:create_HUD_interface()
	-- Create buttons
	for _, v in ipairs(self.buttondefs) do
		db:add_button(v)
	end

	self.cbox = iup.cbox(self.buttons)

	iup.Append(HUD.pluginlayer, iup.vbox{self.cbox})
	HUD.pluginlayer:map()
end

function db:destroy_button_regions()
	local regions = self.regions
	for id, v in pairs(regions) do 
		if (type(id) == 'number' and type(v) == 'number') then 
			rdestroy(id)
			self.regions[id] = nil
		end
	end
end

function db:destroy_field_regions()
	local regions = self.regions
	for id, v in pairs(regions) do 
		if (type(id) == 'number' and v == 'field') then 
			rdestroy(id)
			self.regions[id] = nil
		end
	end
end

function db:destroy_touch_to_target_region()
	local regions = self.regions
	for id, v in pairs(regions) do 
		-- I suppose I could put this into a special case variable, but that would require adding extra logic to the TOUCH events
		-- And those work... I think.
		if (type(id) == 'number' and v == 'target') then 
			rdestroy(id)
			self.regions[id] = nil
			break
		end
	end
end

function db:destroy_touch_to_place_regions()
	local regions = self.regions
	for id, v in pairs(regions) do 
		if (type(id) == 'number' and type(v) == 'function') then 
			rdestroy(id)
			self.regions[id] = nil
		end
	end
end

function db:destroy_all_regions()
	local regions = self.regions
	for id, v in pairs(regions) do 
		if (type(id) == 'number') then 
			rdestroy(id)
			self.regions[id] = nil
		end
	end
end

function db:create_touch_to_target_region()
		self:destroy_touch_to_target_region()
		
		local r = rcreate(nil, nil, nil, false, false, false, false, false, false, 1, 1, self.xres, self.yres, 1) -- TODO change the lower bounds back to 0 after testing
		self.regions[r] = 'target'
end

function db:create_button_regions()

	if (self.disable_default_ui and self.enable_touch_to_target) then self:create_touch_to_target_region() end

	self:destroy_button_regions()

	local r
	for i, v in ipairs(self.buttondefs) do
		if (v.type == 'button') then
			if (v.command1 or v.command2) then -- Only create the region if there's actually a command associated with it
				if (ispluscommand(v.command1)) then
					-- _bind_ +commands to their regions, no long-press option available
					r = rcreate(nil, nil, v.command1 or nil, false, false, false, false, false, false, v.x, v.y, (v.x + v.w - 1), (v.y + v.h - 1))
					self.regions[r] = false -- see comment below for reasoning.  assign false so that it won't be processed in the touch events.
				else
					r = rcreate(nil, nil, nil, false, false, false, false, false, false, v.x, v.y, (v.x + v.w - 1), (v.y + v.h - 1))
					-- Since the return value of CreateTouchRegion is an auto-incrementing number, assign the return value as 
					-- an index with the value of the button it represents for access to the creation values and command, and for destroying it.
					-- Only necessary if the command is not bound to the region.
					self.regions[r] = i
				end
			end
		end
	end
end

function db:create_field_regions()
	if (hud_type) then 
		self:destroy_field_regions() -- probably unnecessary
	
		for i, v in ipairs(self.buttondefs) do
			local r
			if (v.type == 'field') then
				if (hud_type == 'ship') then
					r = rcreate(v.command1, v.command2, v.command3, v.invertX, v.invertY, v.mouselookX, v.mouselookY, v.relative, v.nonlinear, v.x, v.y, (v.x + v.w - 1), (v.y + v.h - 1))
				elseif (hud_type == 'turret') then
					r = rcreate('+LookX', '+LookY', '+LookPress', v.invertX, v.invertY, v.mouselookX, v.mouselookY, v.relative, v.nonlinear, v.x, v.y, (v.x + v.w - 1), (v.y + v.h - 1))
				end
				self.regions[r] = 'field' -- see above (previous value: false)
			end
		end
	end
end

function db:touch_to_target(pW, pH, region)
	local posX = self.xres*pW
	local posY = self.yres*pH
	radar.SetRadarSelectionByPoint(pW, pH, db.touch_to_target_command, region)
end

function db:on_touch_pressed(id, percentW, percentH)
	local buttondef, timer, command2
	local buttonid = self.regions[id]

	-- An entry is only made in regions if command1 is not a +command
	if (type(buttonid) == 'number') then 
		buttondef = self.buttondefs[buttonid] 
		if (buttondef) then 
			timer = self.buttons[buttonid].timer
			if (timer) then -- A timer is not created for short-press +commands.
				--Initiate timer to test for long-press condition.  
				-- If the long-press command is a +command, have it start as soon the timer ends.
				timer:SetTimeout(db.long_press_delay, function()
					command2 = buttondef.command2
					--if command2 is a +command, start executing it.
					if (ispluscommand(command2)) then
						gkinterface.GKProcessCommand(command2)
					end
				end)
			end
		end
	end
end

function db:on_touch_released(id, percentW, percentH)
	local buttondef, timer, command
	local buttonid = self.regions[id]

	-- A non-false entry is only made in regions if command1 is not a +command
	
	-- Region is associated with a button
	if (type(buttonid) == 'number') then 
		buttondef = self.buttondefs[buttonid] 
		if (buttondef) then 
			timer = self.buttons[buttonid].timer

			if (timer:IsActive()) then -- if less than 1 second has passed, execute command1
				command = buttondef.command1 
			else 
				--fall back to command1 if command2 is undefined
				command = buttondef.command2 
				if (command) then
					if (ispluscommand(command)) then command = buttondef.command2cancel end
				else
					command = buttondef.command1 
				end
			end

			if (command) then
				if (type(command) == 'string') then
					gkinterface.GKProcessCommand(command)
				end
			end

			timer:Kill()
		end

	-- Region is part of the Touch-To-Place interface for buttons
	elseif (type(buttonid) == 'function' and self.place_button_mode) then
		buttonid(percentW, percentH)
		
	-- Region is associated with Touch-To-Target
	elseif (buttonid == 'target') then
		self:touch_to_target(percentW, percentH, id)
	end
end

function db:OnEvent(event, ...)
	if     (event == 'HUD_SHOW') then
		hud_type = ...
		-- TODO only create fields
		self:create_field_regions()

	elseif (event == 'HUD_HIDE') then
		hud_type = nil
		-- TODO only destroy fields
		self:destroy_field_regions()

	elseif (event == 'TOUCH_PRESSED') then
		self:on_touch_pressed(...)

	elseif (event == 'TOUCH_RELEASED') then
		self:on_touch_released(...)

	elseif (event == 'PLAYER_ENTERED_GAME') then
		db:set_accel_commands()
		self:load_buttondefs()
		if (not next(db.buttondefs)) then
			db.buttondefs = {
				{title='db', type='button', command1='db', x=100, y=200, w=64, h=64},
				{title='Reload', type='button', command1='reload', x=170, y=200, w=64, h=64},
			}
		end

		if (db.disable_default_ui) then db.remove_default_ui() end
		db:create_HUD_interface()
		db:create_button_regions()
		
	elseif (event == 'UNLOAD_INTERFACE') then
		-- destroy all regions
		--self:destroy_button_regions()
		--self:destroy_field_regions()
		db:destroy_all_regions()
	end
end

db.ui:build_dialog()

RegisterEvent(db, 'HUD_SHOW')
RegisterEvent(db, 'HUD_HIDE')
RegisterEvent(db, 'TOUCH_PRESSED')
RegisterEvent(db, 'TOUCH_RELEASED')
RegisterEvent(db, 'PLAYER_ENTERED_GAME')
RegisterEvent(db, 'UNLOAD_INTERFACE')

RegisterUserCommand('con', function() gkinterface.GKProcessCommand('consoletoggle') end)
RegisterUserCommand('db', function() db.ui.buttonlist:populate() ShowDialog(db.ui.dialog, iup.CENTER, iup.CENTER) end)
RegisterUserCommand('reload', function() print('\127ffffffReloading...') ReloadInterface() end)

--[[ ------------------------------------------------------------------------------------
with the creation of the GUI for adding buttons, this documentation isn't necessary for adding
buttons, However I will retain it for informational purposes.
-- --------------------------------------------------------------------------------------

to add a button to the HUD, add a table to the the buttondefs table.  
it should contain the following keys: title, command1, command2[, command2cancel], font, x, y, w, h

command1 should be assigned the string value of the ingame command to run, or a function.
if command1 is a +command, command2 will not be used

command2 should be assigned the string value of the ingame command to run, or a function.
command2 is the long-press command (1 second wait)
if command2 is a +command, assign the inverse of that command to command2cancel  
  (eg. command2="+shoot2", command2cancel="+shoot2 0")

a list of commands can be found at http://www.vendetta-online.com/manual/app_commands.html

the button examples below are configured for optimal placement on my HTC Inspire 4G.
-- ------------------------------------------------------------------------------------]]
function db.car(start, count)
	start = start or 0
	count = count or 500
	
	for i = start, start + count do
		rdestroy(i)
	end
end